@@ -4,9 +4,9 @@
 # Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020-2021)
 # -----------------------------------------------------------------------------
 
+import colorsys
 import matplotlib.colors as mplc
 import numpy as np
-import colorsys
 
 
 def clip(x, vmin=0.0, vmax=1.0):
@@ -20,43 +20,13 @@ def set_hlsa(
     s: float = None,
     a: float = None,
     to_hex: bool = False,
-):
+) -> tuple:
     """
     Accepts a `matplotlib` color specification and returns an RGB color
     with the specified HLS values plus an optional alpha
 
-    Example:
 
-        import seaborn as sns
-        from matplotlib.colors import ListedColormap
-        import numpy as np
-
-        color_cycle = [
-            '#1f77b4',
-            '#ff7f0e',
-            '#2ca02c',
-            '#d62728',
-            '#9467bd',
-            '#8c564b',
-            '#e377c2',
-            '#7f7f7f',
-            '#bcbd22',
-            '#17becf'
-        ]
-        all_colors = []
-        for col in color_cycle[:2]:
-            hls = colorsys.rgb_to_hls(*mplc.to_rgb(mplc.to_rgb(col)))
-            sat_vals = (np.linspace(0., 1.0, 20) ** 2) * hls[2]
-            alpha_vals = np.linspace(0.4, 1.0, 20)
-
-            colors = [list(set_hlsa(col, s=s)) for s, a in zip(sat_vals, alpha_vals)]
-            all_colors += colors
-
-        def plot_colors(colors):
-            cmap = ListedColormap(colors)
-            sns.palplot(colors)
-
-        plot_colors(all_colors)
+    .. include:: ./docstring_examples/quantify.visualization.color_utilities.set_hlsa.rst.txt
     """
     rgb = mplc.to_rgb(color)
     hls = colorsys.rgb_to_hls(*mplc.to_rgb(rgb))
